from jarbas_house.devices import GenericDevice, RGBWBulb, RGBBulb, Bulb
from jarbas_house.devices.tplink.kasa import discover_devices, find_host_from_device_name, SmartBulb, SmartPlug, \
    tplink_hsv_to_hsv, hsv_to_tplink_hsv
from jarbas_house.colors import Color


class KasaDevice(GenericDevice):
    def __init__(self, host, name="generic kasa device", raw_data=None):
        raw_data = raw_data or {"name": name, "description": "uses tplink Kasa app"}
        super().__init__(host, name, raw_data)

    @property
    def as_dict(self):
        return {
            "host": self.host,
            "name": self.name,
            "model": self.product_model,
            "device_type": "generic tplink kasa device",
            "device_id": self.device_id,
            "state": self.is_on,
            "raw": self.raw_data
        }

    @property
    def product_model(self):
        return self.raw_data.get("model", "kasa")

    @property
    def device_id(self):
        return self.raw_data.get("deviceId")

    @property
    def name(self):
        return self.raw_data["alias"]


class KasaBulb(KasaDevice, Bulb):

    def __init__(self, host, name="light bulb", raw_data=None):
        super().__init__(host, name, raw_data)
        self._timer = None
        self._bulb = SmartBulb(self.host)

    @property
    def as_dict(self):
        return {
            "host": self.host,
            "name": self.name,
            "model": self.product_model,
            "device_type": "light bulb",
            "device_id": self.device_id,
            "color": self.color.as_dict,
            "is_color": self.is_color,
            "is_dimmable": self.is_dimmable,
            "is_variable_color_temp": self.is_variable_color_temp,
            "brightness": self.brightness_255,
            "state": self.is_on,
            "raw": self.raw_data
        }

    @property
    def color(self):
        if self.is_off:
            return Color.from_name("black")
        if self._bulb.is_color:
            h, s, v = tplink_hsv_to_hsv(*self._bulb.hsv)
            return Color.from_hsv(h, s, v)
        return Color.from_name("white")

    @property
    def light_state(self):
        return self.raw_data["light_state"]

    @property
    def is_color(self):
        return self._bulb.is_color

    @property
    def is_dimmable(self):
        return self._bulb.is_dimmable

    @property
    def is_variable_color_temp(self):
        return self._bulb.is_variable_color_temp

    @property
    def is_on(self):
        return self._bulb.is_on

    @property
    def brightness(self):
        return self._bulb.brightness

    @property
    def color_temperatures(self):
        return self._bulb.color_temp

    @property
    def current_consumption(self):
        return self._bulb.get_emeter_realtime()

    @property
    def daily_consumption(self, year=None, month=None):
        return self._bulb.get_emeter_daily(year=year, month=month)

    @property
    def monthly_consumption(self, year=None):
        return self._bulb.get_emeter_monthly(year=year)

    # status change
    def turn_on(self):
        self._bulb.turn_on()

    def turn_off(self):
        self._bulb.turn_off()

    def change_brightness(self, value, percent=True):
        if not percent:
            raise NotImplementedError
        if self._bulb.is_dimmable:
            self._bulb.brightness = value

    def change_color_temperatures(self, value, percent=True):
        if not percent:
            raise NotImplementedError
        if self._bulb.is_variable_color_temp:
            self._bulb.color_temp = value

    def change_color(self, name):
        if isinstance(name, Color):
            if name.rgb == (0, 0, 0):
                self.turn_off()
            else:
                if self.is_off:
                    self.turn_on()
                if self._bulb.is_color:
                    self._bulb.hsv = hsv_to_tplink_hsv(*name.hsv)
        else:
            color = Color.from_name(name)
            self.change_color(color)


class KasaRGBBulb(KasaBulb, RGBBulb):

    def __init__(self, host, name="rgb light bulb", raw_data=None):
        super().__init__(host, name, raw_data)

    @property
    def as_dict(self):
        return {
            "host": self.host,
            "name": self.name,
            "model": self.product_model,
            "device_type": "rgb light bulb",
            "device_id": self.device_id,
            "color": self.color.as_dict,
            "is_color": self.is_color,
            "is_dimmable": self.is_dimmable,
            "is_variable_color_temp": self.is_variable_color_temp,
            "brightness": self.brightness_255,
            "state": self.is_on,
            "raw": self.raw_data
        }


class KasaRGBWBulb(KasaRGBBulb, RGBWBulb):

    def __init__(self, host, name="rgbw light bulb", raw_data=None):
        super().__init__(host, name, raw_data)

    @property
    def as_dict(self):
        return {
            "host": self.host,
            "name": self.name,
            "model": self.product_model,
            "device_type": "rgbw light bulb",
            "device_id": self.device_id,
            "color": self.color.as_dict,
            "is_color": self.is_color,
            "is_dimmable": self.is_dimmable,
            "is_variable_color_temp": self.is_variable_color_temp,
            "brightness": self.brightness_255,
            "state": self.is_on,
            "raw": self.raw_data
        }


def scan_kasa():
    for d in discover_devices():
        if isinstance(d, SmartBulb):
            if d.is_color:
                yield KasaRGBWBulb(d.host, d.alias, dict(d.sys_info))
            else:
                yield KasaBulb(d.host, d.alias, dict(d.sys_info))
        elif isinstance(d, SmartPlug):
            pass


def get_device(ip):
    for device in scan_kasa():
        if device.host == ip:
            return device
    return None


if __name__ == '__main__':
    from pprint import pprint
    from time import sleep

    for device in scan_kasa():
        pprint(device.as_dict)
        device.toggle()
        exit()
        # print(device.is_off, device.is_on)
        # device.turn_on()
        # device.set_high_brightness()
        device.change_color("cyan")
        # device.blink()
        # device.red_blue_cross_fade()
        sleep(30)
        device.reset()

    # bulb = get_device('192.168.1.64')
    # bulb.color_cycle()
    # sleep(60)
    # bulb.reset()



from pyHS100 import Discover, SmartPlug, SmartBulb
from jarbas_house.colors import hex_to_hsv, name_to_hsv, hsv_to_name
from time import sleep
from pyHS100.smartdevice import SmartDeviceException


def discover_devices():
    devices = []
    for ip, dev in Discover.discover().items():
        devices.append(dev)
    return devices


def find_host_from_device_name(devicename, timeout=3, attempts=3,
                               return_dev=False):
    """Discover devices in the network."""
    print("Trying to discover %s using %s attempts of %s seconds" %
             (devicename, attempts, timeout))
    for attempt in range(1, attempts):
        print("Attempt %s of %s" % (attempt, attempts))
        found_devs = Discover.discover(timeout=timeout).items()
        for ip, dev in found_devs:
            if dev.alias.lower() == devicename.lower():
                if return_dev:
                    return dev
                host = dev.host
                return host
    return None


# smart plugs


def get_plug_hw_info(ip=None, device=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        plug = SmartPlug(ip)
    else:
        plug = device
    return plug.hw_info


def get_plug_sys_info(ip=None, device=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        plug = SmartPlug(ip)
    else:
        plug = device
    return plug.get_sysinfo()


def plug_turn_off(ip=None, device=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        plug = SmartPlug(ip)
    else:
        plug = device
    return plug.turn_off()


def plug_turn_on(ip=None, device=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        plug = SmartPlug(ip)
    else:
        plug = device
    return plug.turn_on()


def get_plug_state(ip=None, device=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        plug = SmartPlug(ip)
    else:
        plug = device
    return plug.state


def get_plug_current_consumption(ip=None, device=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        plug = SmartPlug(ip)
    else:
        plug = device
    return plug.get_emeter_realtime()


def get_plug_daily_consumption(ip=None, device=None, year=None, month=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        plug = SmartPlug(ip)
    else:
        plug = device
    return plug.get_emeter_daily(year=year, month=month)


def get_plug_monthly_consumption(ip=None, device=None, year=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        plug = SmartPlug(ip)
    else:
        plug = device
    return plug.get_emeter_monthly(year=year)


def plug_led(ip=None, device=None, state=True):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        plug = SmartPlug(ip)
    else:
        plug = device
    plug.led = state


def plug_led_on(ip=None, device=None):
    plug_led(ip, device, True)


def plug_led_off(ip=None, device=None):
    plug_led(ip, device, False)


def plug_reboot(ip=None, device=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        plug = SmartPlug(ip)
    else:
        plug = device
    # TODO Turn off, check state, when off, turn on maybe even with an
    # optional time to off?
    plug_turn_off(device=plug)
    # state = get_plug_state(device=plug)
    sleep(1)
    plug_turn_on(device=plug)


# smart bulbs


def get_bulb_hw_info(ip=None, device=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        bulb = SmartBulb(ip)
    else:
        bulb = device
    return bulb.hw_info


def get_bulb_sys_info(ip=None, device=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        bulb = SmartBulb(ip)
    else:
        bulb = device
    return bulb.get_sysinfo()


def bulb_turn_off(ip=None, device=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        bulb = SmartBulb(ip)
    else:
        bulb = device
    return bulb.turn_off()


def bulb_turn_on(ip=None, device=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        bulb = SmartBulb(ip)
    else:
        bulb = device
    return bulb.turn_on()


def get_bulb_state(ip=None, device=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        bulb = SmartBulb(ip)
    else:
        bulb = device
    return bulb.state


def get_bulb_current_consumption(ip=None, device=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        bulb = SmartBulb(ip)
    else:
        bulb = device
    return bulb.get_emeter_realtime()


def get_bulb_daily_consumption(ip=None, device=None, year=None, month=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        bulb = SmartBulb(ip)
    else:
        bulb = device
    return bulb.get_emeter_daily(year=year, month=month)


def get_bulb_monthly_consumption(ip=None, device=None, year=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        bulb = SmartBulb(ip)
    else:
        bulb = device
    return bulb.get_emeter_monthly(year=year)


def set_bulb_brightness(ip=None, device=None, percentage=100):
    percentage = int(percentage)
    if percentage > 100:
        percentage = 100
    if percentage < 0:
        percentage = 0

    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        bulb = SmartBulb(ip)
    else:
        bulb = device
    if bulb.is_dimmable:
        bulb.brightness = percentage


def get_bulb_brightness(ip=None, device=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        bulb = SmartBulb(ip)
    else:
        bulb = device
    return bulb.brightness


def set_bulb_color_temperature(ip=None, device=None, value=3000):
    value = int(value)
    if value > 100:
        value = 100
    if value < 0:
        value = 0

    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        bulb = SmartBulb(ip)
    else:
        bulb = device
    if bulb.is_variable_color_temp:
        bulb.color_temp = value


def get_bulb_color_temperature(ip=None, device=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        bulb = SmartBulb(ip)
    else:
        bulb = device
    return bulb.color_temp


def set_bulb_color(ip=None, device=None, hex_color=None, color_name=None):
    if hex_color is None and color_name is None:
        raise AttributeError("no color specified")
    if hex_color is None:
        h, s, v = name_to_hsv(color_name)
    else:
        h, s, v = hex_to_hsv(hex_color)
    set_bulb_hsv(ip, device, h, s, v)
    return color_name


def set_bulb_hsv(ip=None, device=None, hue=0.5, saturation=1, value=255):
    if ip is None and device is None:
        raise AttributeError("no device specified")

    hue, saturation, value = hsv_to_tplink_hsv(hue, saturation, value)

    if device is None:
        bulb = SmartBulb(ip)
    else:
        bulb = device
    if bulb.is_color:
        bulb.hsv = (hue, saturation, value)


def tplink_hsv_to_hsv(h, s, v):
    # 0 < h < 360 <- tplink
    # 0 < h < 1 <- colorsys hsv
    h = h / 360
    # 0 < h < 100 <- tplink
    # 0 < h < 1 <- colorsys
    s = s / 100
    # 0 < v < 100 <- tplink
    # 0 < v < 255 <- colorsys hsv
    v = v / 100 * 255
    return h, s, int(v)


def hsv_to_tplink_hsv(h, s, v):
    # 0 < h < 360 <- tplink
    # 0 < h < 1 <- colorsys hsv
    h = h * 360
    # 0 < h < 100 <- tplink
    # 0 < h < 1 <- colorsys
    s = s * 100
    # 0 < v < 100 <- tplink
    # 0 < v < 255 <- colorsys hsv
    v = v * 100 / 255
    return int(h), int(s), int(v)


def get_bulb_hsv(ip=None, device=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        bulb = SmartBulb(ip)
    else:
        bulb = device
    if bulb.is_color:
        return bulb.hsv


def get_bulb_color_name(ip=None, device=None):
    if ip is None and device is None:
        raise AttributeError("no device specified")
    if device is None:
        bulb = SmartBulb(ip)
    else:
        bulb = device
    if bulb.is_color:
        h, s, v = bulb.hsv
        return hsv_to_name(h, s, v)
    return "unknown color"



from threading import Thread


def merge_dict(base, delta):
    """
        Recursively merging configuration dictionaries.

        Args:
            base:  Target for merge
            delta: Dictionary to merge into base
    """

    for k, dv in delta.items():
        bv = base.get(k)
        if isinstance(dv, dict) and isinstance(bv, dict):
            merge_dict(bv, dv)
        else:
            base[k] = dv


def create_daemon(func):
    t = Thread(target=func)
    t.setDaemon(True)
    t.start()
    return t



from jarbas_house.devices import GenericDevice
from jarbas_house.util import merge_dict
import time
from threading import Timer


class House:
    def __init__(self, name="house", scanners=None, verbose=True):
        self._name = name
        self.scanners = scanners or []
        self._devices = {}
        self._timestamps = {}
        self._device_timeout = 3 * 60
        self._monitor_pause = 10
        self.verbose = verbose
        self._timer = None
        self.monitoring = False
        self._on_new_device_handlers = []
        self._on_device_seen_handlers = []
        self._on_device_updated_handlers = []
        self._on_device_lost_handlers = []

    def add_scanner(self, func):
        self.scanners.append(func)

    def add_device(self, device):
        assert isinstance(device, GenericDevice)
        self.devices[device.host] = device

    def remove_device(self, device):
        if isinstance(device, GenericDevice):
            host = device.host
        else:
            host = str(device)
        self.devices.pop(host)
        self._timestamps.pop(host)

    def update_device(self, device):
        assert isinstance(device, GenericDevice)
        # TODO make this part of device class
        if device.raw_data not in self._devices[device.host]._raw:
            self._devices[device.host]._raw.append(device.raw_data)
            if self.verbose:
                print("DEBUG: updating device data {device}".format(device=self._devices[device.host]))
            for handler in self._on_device_updated_handlers:
                try:
                    handler(device)
                except Exception as e:
                    if self.verbose:
                        print("ERROR: exception in on_device_updated handler - {e}".format(e=e))

            if device.name:
                self._devices[device.host]._name = device.name
            # TODO implement an hierarchy for class morphing
            if self._devices[device.host].__class__ == GenericDevice and device.__class__ != GenericDevice:
                if self.verbose:
                    print("DEBUG: updating device class {device}".format(device=self._devices[device.host]))
                self._devices[device.host].__class__ = device.__class__

    @property
    def devices(self):
        return self._devices

    @property
    def name(self):
        return self._name

    def scan_devices(self):
        for scanner in self.scanners:
            for device in scanner():
                yield device

    def start(self):
        self.monitoring = True
        self._timer = Timer(self._monitor_pause, self._monitor)
        self._timer.setDaemon(True)
        self._timer.start()

    def stop(self):
        self.monitoring = False
        self._timer = None

    def _monitor(self):
        if self.monitoring:
            for device in self.scan_devices():
                if device.host not in self._devices:
                    if self.verbose:
                        print("INFO: new device:{device}".format(device=device))
                    for handler in self._on_new_device_handlers:
                        try:
                            handler(device)
                        except Exception as e:
                            if self.verbose:
                                print("ERROR: exception in on_new_device handler - {e}".format(e=e))
                    self.add_device(device)
                else:
                    for handler in self._on_device_seen_handlers:
                        try:
                            handler(device)
                        except Exception as e:
                            if self.verbose:
                                print("ERROR: exception in on_device_seen handler - {e}".format(e=e))
                    self.update_device(device)

                if device.host not in self._timestamps:
                    self._timestamps[device.host] = []
                self._timestamps[device.host] += [time.time()]

            now = time.time()
            for device in dict(self._timestamps):
                last_seen = now - self._timestamps[device][-1]
                if self.verbose:
                    print(
                        "DEBUG: device {device}, last seen {seconds} seconds ago ".format(device=self._devices[device],
                                                                                          seconds=last_seen))
                if last_seen >= self._device_timeout or not self._devices[device].is_online:
                    if self.verbose:
                        print("INFO: lost device: {device}".format(device=self._devices[device]))
                    for handler in self._on_device_lost_handlers:
                        try:
                            handler(device)
                        except Exception as e:
                            if self.verbose:
                                print("ERROR: exception in on_device_lost handler - {e}".format(e=e))
                    self.remove_device(device)

            self.start()

    def on_new_device(self, handler):
        self._on_new_device_handlers.append(handler)

    def on_device_lost(self, handler):
        self._on_device_lost_handlers.append(handler)

    def on_device_seen(self, handler):
        self._on_device_seen_handlers.append(handler)

    def on_device_updated(self, handler):
        self._on_device_updated_handlers.append(handler)

    def remove_handlers(self):
        self._on_new_device_handlers = []
        self._on_device_seen_handlers = []
        self._on_device_lost_handlers = []
        self._on_device_updated_handlers = []


if __name__ == "__main__":
    from pprint import pprint
    from jarbas_house.devices.ewelink import scan_ewelink
    from jarbas_house.devices.tplink import scan_kasa
    from jarbas_house.devices.magic_home import scan_magichome
    from jarbas_house.devices.scan import scan_bluetooth, scan_wifi

    house = House()
    house.add_scanner(scan_wifi)
    house.add_scanner(scan_bluetooth)
    house.add_scanner(scan_magichome)
    house.add_scanner(scan_kasa)
    house.add_scanner(scan_ewelink)

    # for d in house.scan_devices():
    #    pprint(d.as_dict)

    # house = House()

    house.start()
    try:
        while True:
            time.sleep(5)
    except KeyboardInterrupt:
        house.stop()



from setuptools import setup

setup(
    name='jarbas_house',
    version='0.1',
    packages=['jarbas_house', 'jarbas_house.colors', 'jarbas_house.devices', 'jarbas_house.devices.tplink',
              'jarbas_house.devices.ewelink', 'jarbas_house.devices.magic_home'],
    url='',
    install_requires=[
        "webcolors",
        "pyHS100",
        "requests",
        "websocket-client==0.54.0",
        "python-nmap"
    ],
    license='MIT',
    author='jarbasai',
    author_email='jarbasai@mailfence.com',
    description='tools to control my iot devices'
)
